#include "shoki_includes.h"
#include "shoki.h"
#include "shoki_doctrine.h"
#include "shoki_callback.h"
#include "shoki_common_utils.h"

#ifndef SHOKI_LOG_NAME
#define SHOKI_LOG_NAME	"shoki"
#endif

extern char *currentfile, *dev, *fname, *outfile, *filterlist;
extern char *ostype, *portlist, **filter_pt;
extern struct bpf_program **fcode_pt;
extern int count, default_snaplen, start, stop;
extern int query_window, query_max, query_period, nfilters;
extern pcap_t *pd;
extern struct bpf_program base_fcode;
char email_rcpt[256];
char *fprint_file;


static struct { const char *name; int option; } option[] =
{
	{ "count",		OPT_COUNT },
	{ "current",		OPT_CURRENT },
	{ "dev",		OPT_DEV },
	{ "dumps",		OPT_DUMPS },
	{ "email",		OPT_EMAIL },
	{ "filter",		OPT_FILTER },
	{ "filterlist",		OPT_FILTERLIST },
	{ "snaplen",		OPT_SNAPLEN },
	{ "start",		OPT_START },
	{ "stop",		OPT_STOP },
	{ "query_window",	OPT_WINDOW },
	{ "query_max",		OPT_MAX },
	{ "query_period",	OPT_PERIOD },
	{ "ports",		OPT_PORTS },
	{ "ostype",		OPT_OSTYPE },
	{ "fingerprints",	OPT_FPRINTS },
	{ "format",		OPT_FORMAT },
	{ NULL,		0 }
};

static struct { const char *name; int option; } output_list[] =
{
	{ "SRC",	PF_SRC },
	{ "SPORT",	PF_SPORT },
	{ "DST",	PF_DST },
	{ "DPORT",	PF_DPORT },
	{ "PROTO",	PF_PROTO },
	{ "SEQ",	PF_SEQ },
	{ "DF",		PF_DF },
	{ "SYN",	PF_SYN },
	{ "FIN",	PF_FIN },
	{ "RST",	PF_RST },
	{ "PUSH",	PF_PUSH },
	{ "TIME",	PF_TIME },
	{ "ICMP_TYPE",	PF_ICMP_TYPE },
	{ "ACK",	PF_ACK },
	{ "LEN",	PF_LEN },
	{ "FLAGS",	PF_FLAGS },
	{ "NAME",	PF_NAME },
	{ "ID",		PF_ID },
	{ "DATA",	PF_DATA },
	{ "HEX",	PF_HEX },
	{ "SEC",	PF_SEC },
	{ "USEC",	PF_USEC },
	{ "VER",	PF_VER },
	{ "TOS",	PF_TOS },
	{ "FRAG",	PF_FRAG },
	{ "TTL",	PF_TTL },
	{ "IPSUM",	PF_IPSUM },
	{ "WIN",	PF_WIN },
	{ "URG",	PF_URG },
	{ "IPID",	PF_IPID },
	{ "SENSOR",	PF_SENSOR },
	{ "ALERT",	PF_ALERT },
	{ "RUNTIME",	PF_RUNTIME },
	{ NULL,		0 }
};

char
*read_file(char *fname)
{
	register int f, bar;
	register char *foo;
	int i;

	struct stat stat;
	f = open(fname, O_RDONLY);
	if(f < 0) printf("Couldn't read %s\n", fname), exit(1);
	if(fstat(f, &stat) < 0) printf("Couldn't stat %s\n", fname), exit(1);
	foo = malloc((u_int)stat.st_size + 1);
	bar = read(f, foo, (int)stat.st_size);
	if(bar < 0) printf("Problem reading %s\n", fname), exit(1);
	if(bar != stat.st_size)
		printf("short read reading %s\n", fname), exit(1);
	foo[(int)stat.st_size] = '\0';
	i = (int)stat.st_size - 1;
	while((i >= 0) && isspace(foo[i]))
		foo[i--] = '\0';
	return(foo);
}

void
error(const char *fmt, ...)
{
	va_list va;
	char buf[512];

	memset(buf, '\0', 512);
	va_start(va, fmt);
	vsnprintf(buf, sizeof(buf), fmt, va);
	va_end(va);
	fprintf(stderr, "%s\n", buf);
	fflush(stderr);
	(void)openlog(SHOKI_LOG_NAME, LOG_PID, SHOKI_LOG_SVC);
        syslog(SHOKI_LOG_LVL, "%s", buf);
        (void)closelog();
	exit(1);
}

#define WHITE "= \t\n\r"

void
parse_shoki_config(const char *config)
{
	FILE *f = NULL;
	char line[1024];
	int i, j, k, l, m = 0, match = -1;
	char *foo, *bar, *baz;
	int tmp_pf_f[MAX_FORMAT_ARGS];
	char *tmp_pf_d[MAX_FORMAT_ARGS];

	f = fopen(config, "r");
	if(!f)	error("Couldn't read %s", config);
	while(fgets(line, sizeof(line), f))
	{
		foo = line + strspn(line, WHITE);
		if(*foo == '#' || !*foo) continue;
		if(!(bar = strdup(foo)))
			error("strdup:  out of memory in parse_shoki_config");
		foo = strtok(foo, WHITE);	/* I hate myself */
		for(i=0; foo[i]; i++)	foo[i] = tolower(foo[i]);
		match = -1;
		for(j=0; option[j].name && match == -1; j++)
		{
			if(strcmp(foo, option[j].name) == 0)
				match = option[j].option;
		}
		if(match == -1) error("Bad config option");
		baz = bar + strspn(bar, foo);
		baz = baz + strspn(baz, WHITE);
		foo = strtok(NULL, WHITE);
		if(!*foo) error("No argument for config option");
		switch (match)
		{
			case OPT_COUNT:
				if(count == -1) count = atoi(foo);
				break;
			case OPT_CURRENT:
				if(currentfile == NULL)
					if(!(currentfile = strdup(foo)))
						error("strdup:  out of memory in parse_shoki_config");
				break;
			case OPT_DEV:
				if(dev == NULL)
					if(!(dev = strdup(foo)))
						error("strdup:  out of memory in parse_shoki_config");
				break;
			case OPT_DUMPS:
				if(outfile == NULL)
					if(!(outfile = strdup(foo)))
						error("strdup:  out of memory in parse_shoki_config");
				break;
			case OPT_EMAIL:
				if(email_rcpt == NULL)
					snprintf(email_rcpt, 256, "%s", foo);
				break;
			case OPT_FILTER:
				if(fname == NULL)
					if(!(fname = strdup(foo)))
						error("strdup:  out of memory in parse_shoki_config");
				break;
			case OPT_FILTERLIST:
				if(filterlist == NULL)
					if(!(filterlist = strdup(foo)))
						error("strdup:  out of memory in parse_shoki_config");
				break;
			case OPT_SNAPLEN:
				if(default_snaplen == -1)
					default_snaplen = atoi(foo);
				break;
			case OPT_START:
				if(start == 0) start = atoi(foo);
				break;
			case OPT_STOP:
				if(stop == 0) stop = atoi(foo);
				break;
			case OPT_WINDOW:
				if(query_window == 0)
					query_window = atoi(foo);
				break;
			case OPT_MAX:
				if(query_max == 0) query_max = atoi(foo);
				break;
			case OPT_PERIOD:
				if(query_period == 0)
					query_period = atoi(foo);
				break;
			case OPT_PORTS:
				if(portlist == NULL)
					if(!(portlist = strdup(foo)))
						error("strdup:  out of memory in parse_shoki_config");
				break;
			case OPT_OSTYPE:
				if(ostype == NULL)
				{
					bar = baz;
					while(*bar && *bar != '\n' && *bar != '#') bar++;
					while(isspace((int)*bar)) bar--;
					bar++;
					*bar = '\0';
					if(!(ostype = strdup(baz)))
						error("strdup:  out of memory in parse_shoki_config");
				}
				break;
			case OPT_FPRINTS:
				if(fprint_file == NULL)
				{
					if(!(fprint_file = strdup(foo)))
						error("strdup:  out of memory in parse_shoki_config");
				}
				break;
			case OPT_FORMAT:
				/* Abandon all hope, ye who enter here */
				bar = baz;
				while(*bar && *bar != '\n' && *bar != '#')
					bar++;
				while(isspace((int)*bar)) bar--;
				bar++;
				*bar = '\0';
				m = 0;
				k = 0;
				l = 0;
				while((bar = rindex(baz, '%')) && (bar != NULL))
				{
					if(m > MAX_FORMAT_ARGS)
						error("Too many format args in config file");
					*bar = '\0';
					bar++;
					i = 0;
					j = 0;
					while((output_list[i].name != NULL) &&
						(j == 0))
					{
						if((strncmp(bar, output_list[i].name, strlen(bar)) == 0) && (j == 0))
							{
								tmp_pf_f[k] = output_list[i].option;
								/* WHAT?
								if(!(tmp_pf_f[k] = strdup(output_list[i].option)))
									error("strdup:  out of memory in parse_shoki_config");
								*/
									
								pf_type[m] = PF_TYPE_VAR;
								k++;
								j++;
								m++;
							}
						i++;
					}
					if(j == 0)
					{
						if(!(tmp_pf_d[l] = strdup(bar)))
							error("strdup:  out of memory in parse_shoki_config");
						pf_type[m] = PF_TYPE_DELIM;
						l++;
						m++;
					}
				}
				if(baz != NULL)
				{
					i = 0;
					j = 0;
					while((output_list[i].name != NULL) &&
						(j == 0))
					{
						if((strncmp(baz, output_list[i].name, strlen(baz)) == 0) && (j == 0))
							{
								tmp_pf_f[k] = output_list[i].option;
								pf_type[m] = PF_TYPE_VAR;
								k++;
								j++;
								m++;
							}
						i++;
					}
					if(j == 0)
					{
						/*
						if(!(tmp_pf_d[l] = (char *)calloc(1, strlen(baz))))
							error("calloc:  out of memory in packet_output");
						snprintf(tmp_pf_d[l], strlen(baz), "%s", baz);
						*/
						if(!(tmp_pf_d[l] = strdup(baz)))
							error("strdup:  out of memory in packet_output");
						pf_type[m] = PF_TYPE_DELIM;
						l++;
						m++;
					}
				}
				for(i = (k - 1), j = 0; i >= 0; i--, j++)
				{
					pf_outformat[j] = tmp_pf_f[i];
				}
				for(i = (l - 1), j = 0; i >= 0; i--, j++)
				{
					if(!(pf_outdelim[j] = strdup(tmp_pf_d[i])))
						error("strdup:  out of memory in packet_output");
				}
				pf_type[m] = -1;
				break;
			default:
				error("Bad config option (%d)", match);
				break;
		}

	}
}

